@@ -64,8 +64,32 @@ class DenonAVRAudyssey(DenonAVRFoundation):
         for tag in self.appcommand0300_attrs:
             self._device.api.add_appcommand0300_update_tag(tag)
 
+        self._device.telnet_api.register_callback(
+            "PS", self._sound_detail_callback
+        )
+
         self._is_setup = True
 
+    async def _sound_detail_callback(
+            self,
+            zone: str,
+            event: str,
+            parameter: str) -> None:
+        """Handle a sound detail change event."""
+        if self._device.zone != zone:
+            return
+
+        if parameter[0:6] == "REFLEV":
+            self._reflevoffset = parameter[7:]
+        elif parameter[0:6] == "DYNVOL":
+            self._dynamicvol = parameter[7:]
+        elif parameter[0:6] == "MULTEQ":
+            self._multeq = parameter[7:]
+        elif parameter == "DYNEQ ON":
+            self._dynamiceq = "1"
+        elif parameter == "DYNEQ OFF":
+            self._dynamiceq = "0"
+
     async def async_update(
             self,
             global_update: bool = False,
