@@ -73,13 +73,15 @@ TELNET_SOURCES = ["CD", "PHONO", "TUNER", "DVD", "BD", "TV",
 ALL_ZONE_STEREO = "ALL ZONE STEREO"
 
 SOUND_MODE_MAPPING = {
-    "MUSIC": ["PLII MUSIC", "DTS NEO:6 MUSIC", "DOLBY D +NEO:X M",
-              "DTS NEO:X MUSIC", "DOLBY PL2 MUSIC", "PLIIX MUSIC"],
+    "MUSIC": ["PLII MUSIC", "DTS NEO:6 MUSIC", "DTS NEO:6 M",
+              "DOLBY D +NEO:X M", "DTS NEO:X MUSIC", "DOLBY PL2 MUSIC",
+              "DOLBY PL2 M", "PLIIX MUSIC", "DOLBY PL2 X MUSIC"],
     "MOVIE": ["PLII MOVIE", "PLII CINEMA", "DTS NEO:X CINEMA",
-              "DTS NEO:6 CINEMA", "DOLBY D +NEO:X C",
+              "DTS NEO:6 CINEMA", "DTS NEO:6 C", "DOLBY D +NEO:X C",
               "PLIIX CINEMA", "DOLBY PLII MOVIE", "MULTI IN + VIRTUAL:X",
-              "DOLBY PL2 CINEMA"],
-    "GAME": ["PLII GAME", "DOLBY D +NEO:X G", "DOLBY PL2 GAME"],
+              "DOLBY PL2 CINEMA", "DOLBY PL2 C", "DOLBY PL2 X MOVIE"],
+    "GAME": ["PLII GAME", "DOLBY D +NEO:X G", "DOLBY PL2 GAME", "DOLBY PL2 G",
+             "DOLBY PL2 X GAME"],
     "AUTO": ["None"],
     "STANDARD": ["None2"],
     "VIRTUAL": ["VIRTUAL"],
