@@ -7,18 +7,18 @@ This module implements the interface to Denon AVR receivers.
 :license: MIT, see LICENSE for more details.
 """
 
+import asyncio
 import logging
 import time
 
-from typing import Callable, Dict, List, Optional
+from typing import Awaitable, Callable, Dict, List, Optional
 
 import attr
 import httpx
 
 from .decorators import run_async_synchronously
 from .foundation import DenonAVRFoundation, set_api_host, set_api_timeout
-from .const import (
-    DENON_ATTR_SETATTR, MAIN_ZONE, VALID_ZONES)
+from .const import (DENON_ATTR_SETATTR, MAIN_ZONE, VALID_ZONES)
 from .exceptions import AvrCommandError
 
 from .audyssey import DenonAVRAudyssey, audyssey_factory
@@ -76,6 +76,7 @@ class DenonAVR(DenonAVRFoundation):
             attr.validators.instance_of(dict)),
         default=attr.Factory(dict),
         init=False)
+    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
     audyssey: DenonAVRAudyssey = attr.ib(
         validator=attr.validators.instance_of(DenonAVRAudyssey),
         default=attr.Factory(audyssey_factory, takes_self=True),
@@ -128,19 +129,24 @@ class DenonAVR(DenonAVRFoundation):
 
     async def async_setup(self) -> None:
         """Ensure that configuration is loaded from receiver asynchronously."""
-        # Device setup
-        await self._device.async_setup()
-        if self._name is None:
-            self._name = self._device.friendly_name
+        async with self._setup_lock:
+            # Device setup
+            await self._device.async_setup()
+            if self._name is None:
+                self._name = self._device.friendly_name
 
-        # Setup other functions
-        self.input.setup()
-        await self.soundmode.async_setup()
-        self.tonecontrol.setup()
-        self.vol.setup()
-        self.audyssey.setup()
+            # Setup other functions
+            self.input.setup()
+            await self.soundmode.async_setup()
+            self.tonecontrol.setup()
+            self.vol.setup()
+            self.audyssey.setup()
 
-        self._is_setup = True
+            for zone_name, zone_item in self._zones.items():
+                if zone_name != self.zone:
+                    await zone_item.async_setup()
+
+            self._is_setup = True
 
     @run_async_synchronously(async_func=async_setup)
     def setup(self) -> None:
@@ -215,6 +221,30 @@ class DenonAVR(DenonAVRFoundation):
     def send_get_command(self, request: str) -> str:
         """Send HTTP GET command to Denon AVR receiver...for compatibility."""
 
+    def register_callback(
+        self,
+        event: str,
+        callback: Callable[[str, str, str], Awaitable[None]]
+    ):
+        """Register a callback for telnet events."""
+        self._device.telnet_api.register_callback(event, callback=callback)
+
+    def unregister_callback(
+        self,
+        event: str,
+        callback: Callable[[str, str, str], Awaitable[None]]
+    ):
+        """Unregister a callback for telnet events."""
+        self._device.telnet_api.unregister_callback(event, callback=callback)
+
+    async def async_telnet_connect(self):
+        """Connect to the telnet interface of the receiver."""
+        await self._device.telnet_api.async_connect()
+
+    async def async_telnet_disconnect(self):
+        """Disconnect from the telnet interface of the receiver."""
+        await self._device.telnet_api.async_disconnect()
+
     ##############
     # Properties #
     ##############
@@ -319,7 +349,7 @@ class DenonAVR(DenonAVRFoundation):
         return self.soundmode.sound_mode_list
 
     @property
-    def sound_mode_map(self) -> Dict[str, str]:  # returns an OrderedDict
+    def sound_mode_map(self) -> Dict[str, str]:
         """Return a dict of available sound modes with their mapping values."""
         return self.soundmode.sound_mode_map
 
@@ -400,6 +430,16 @@ class DenonAVR(DenonAVRFoundation):
             return None
         return self._device.receiver.type
 
+    @property
+    def telnet_connected(self) -> bool:
+        """Return True if telnet is connected."""
+        return self._device.telnet_api.connected
+
+    @property
+    def telnet_healthy(self) -> Optional[bool]:
+        """Return True if telnet connection is healthy."""
+        return self._device.telnet_api.healthy
+
     @property
     def show_all_inputs(self) -> Optional[bool]:
         """Indicate if all inputs are shown or just active one."""
