@@ -7,11 +7,11 @@ This module implements the handler for sound mode of Denon AVR receivers.
 :license: MIT, see LICENSE for more details.
 """
 
+import asyncio
 from copy import deepcopy
 import logging
 
-from collections import OrderedDict
-from typing import Dict, Hashable, Optional
+from typing import Dict, Hashable, List, Optional
 
 import attr
 
@@ -25,11 +25,19 @@ from .foundation import DenonAVRFoundation
 _LOGGER = logging.getLogger(__name__)
 
 
-def rstrip_string(value: Optional[str]) -> Optional[str]:
-    """Perform HTML unescape on value."""
+def convert_sound_mode(value: Optional[str]) -> Optional[str]:
+    """Remove multiple spaces from value."""
     if value is None:
         return value
-    return (str(value)).rstrip()
+    return " ".join(str(value).split())
+
+
+def sound_mode_map_factory() -> Dict[str, List]:
+    """Construct sound_mode map."""
+    sound_mode_map = {}
+    for matched_mode, sublist in SOUND_MODE_MAPPING.items():
+        sound_mode_map[matched_mode] = [convert_sound_mode(i) for i in sublist]
+    return sound_mode_map
 
 
 def sound_mode_rev_map_factory(instance: DenonAVRFoundation) -> Dict[str, str]:
@@ -46,7 +54,7 @@ def sound_mode_rev_map_factory(instance: DenonAVRFoundation) -> Dict[str, str]:
     mode_map_rev = {}
     for matched_mode, sublist in mode_map:
         for raw_mode in sublist:
-            mode_map_rev[raw_mode.upper()] = matched_mode
+            mode_map_rev[convert_sound_mode(raw_mode)] = matched_mode
     return mode_map_rev
 
 
@@ -58,17 +66,16 @@ class DenonAVRSoundMode(DenonAVRFoundation):
         converter=attr.converters.optional(bool),
         default=None)
     _sound_mode_raw: Optional[str] = attr.ib(
-        converter=attr.converters.optional(rstrip_string),
+        converter=attr.converters.optional(convert_sound_mode),
         default=None)
-    _sound_mode_map: Dict[str, str] = attr.ib(  # in fact it is an OrderedDict
+    _sound_mode_map: Dict[str, list] = attr.ib(
         validator=attr.validators.deep_mapping(
             attr.validators.instance_of(str),
             attr.validators.deep_iterable(
                 attr.validators.instance_of(str),
                 attr.validators.instance_of(list)),
-            attr.validators.instance_of(OrderedDict)),
-        default=SOUND_MODE_MAPPING,
-        init=False)
+            attr.validators.instance_of(dict)),
+        default=attr.Factory(sound_mode_map_factory), init=False)
     _sound_mode_map_rev: Dict[str, str] = attr.ib(
         validator=attr.validators.deep_mapping(
             attr.validators.instance_of(str),
@@ -76,6 +83,7 @@ class DenonAVRSoundMode(DenonAVRFoundation):
             attr.validators.instance_of(dict)),
         default=attr.Factory(sound_mode_rev_map_factory, takes_self=True),
         init=False)
+    _setup_lock: asyncio.Lock = attr.ib(default=attr.Factory(asyncio.Lock))
 
     # Update tags for attributes
     # AppCommand.xml interface
@@ -89,18 +97,35 @@ class DenonAVRSoundMode(DenonAVRFoundation):
 
     async def async_setup(self) -> None:
         """Ensure that the instance is initialized."""
-        # Add tags for a potential AppCommand.xml update
-        for tag in self.appcommand_attrs:
-            self._device.api.add_appcommand_update_tag(tag)
+        async with self._setup_lock:
+            # Add tags for a potential AppCommand.xml update
+            for tag in self.appcommand_attrs:
+                self._device.api.add_appcommand_update_tag(tag)
 
-        # Soundmode is always available for AVR-X and AVR-X-2016 receivers
-        # For AVR receiver it will be tested druing the first update
-        if self._device.receiver in [AVR_X, AVR_X_2016]:
-            self._support_sound_mode = True
-        else:
-            await self.async_update_sound_mode()
+            # Soundmode is always available for AVR-X and AVR-X-2016 receivers
+            # For AVR receiver it will be tested druing the first update
+            if self._device.receiver in [AVR_X, AVR_X_2016]:
+                self._support_sound_mode = True
+            else:
+                await self.async_update_sound_mode()
 
-        self._is_setup = True
+            self._device.telnet_api.register_callback(
+                "MS",
+                self._soundmode_callback
+            )
+
+            self._is_setup = True
+
+    async def _soundmode_callback(
+            self,
+            zone: str,
+            event: str,
+            parameter: str) -> None:
+        """Handle a sound mode change event."""
+        if self._device.zone != zone:
+            return
+
+        self._sound_mode_raw = parameter
 
     async def async_update(
             self,
@@ -145,20 +170,43 @@ class DenonAVRSoundMode(DenonAVRFoundation):
             raise AvrProcessingError(
                 "Device is not setup correctly, update method not set")
 
-    def match_sound_mode(self, sound_mode_raw: str) -> Optional[str]:
+    def match_sound_mode(self) -> Optional[str]:
         """Match the raw_sound_mode to its corresponding sound_mode."""
         if self._sound_mode_raw is None:
             return None
+        smr_up = self._sound_mode_raw.upper()
         try:
-            sound_mode = self._sound_mode_map_rev[sound_mode_raw.upper()]
-            return sound_mode
+            sound_mode = self._sound_mode_map_rev[smr_up]
         except KeyError:
-            smr_up = sound_mode_raw.upper()
-            self._sound_mode_map[smr_up] = [smr_up]
+            # Estimate sound mode for unclassified input
+            if smr_up.find("DTS") != -1:
+                self._sound_mode_map["DTS SURROUND"].append(smr_up)
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "assuming 'DTS SURROUND'.", smr_up)
+            elif smr_up.find("DOLBY") != -1:
+                self._sound_mode_map["DOLBY DIGITAL"].append(smr_up)
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "assuming 'DOLBY DIGITAL'.", smr_up)
+            elif smr_up.find("MUSIC") != -1:
+                self._sound_mode_map["MUSIC"].append(smr_up)
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "assuming 'MUSIC'.", smr_up)
+            elif smr_up.find("AURO") != -1:
+                self._sound_mode_map["AURO3D"].append(smr_up)
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "assuming 'AURO3D'.", smr_up)
+            elif (smr_up.find("MOVIE") != -1 or smr_up.find("CINEMA") != -1):
+                self._sound_mode_map["MOVIE"].append(smr_up)
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "assuming 'MOVIE'.", smr_up)
+            else:
+                self._sound_mode_map[smr_up] = [smr_up]
+                _LOGGER.warning("Not able to match sound mode: '%s', "
+                                "returning raw sound mode.", smr_up)
             self._sound_mode_map_rev = sound_mode_rev_map_factory(self)
-            _LOGGER.warning("Not able to match sound mode: '%s', "
-                            "returning raw sound mode.", smr_up)
-        return sound_mode_raw
+            sound_mode = self._sound_mode_map_rev[smr_up]
+
+        return sound_mode
 
     async def _async_set_all_zone_stereo(self, zst_on: bool) -> None:
         """
@@ -185,7 +233,7 @@ class DenonAVRSoundMode(DenonAVRFoundation):
     @property
     def sound_mode(self) -> Optional[str]:
         """Return the matched current sound mode as a string."""
-        sound_mode_matched = self.match_sound_mode(self._sound_mode_raw)
+        sound_mode_matched = self.match_sound_mode()
         return sound_mode_matched
 
     @property
@@ -194,7 +242,7 @@ class DenonAVRSoundMode(DenonAVRFoundation):
         return list(self._sound_mode_map.keys())
 
     @property
-    def sound_mode_map(self) -> Dict[str, str]:  # returns an OrderedDict
+    def sound_mode_map(self) -> Dict[str, str]:
         """Return a dict of available sound modes with their mapping values."""
         return deepcopy(self._sound_mode_map)
 
