@@ -2,7 +2,6 @@
 # -*- coding: utf-8 -*-
 """Tests for ssdp."""
 
-import requests_mock
 import pytest
 from denonavr.ssdp import evaluate_scpd_xml
 
@@ -22,7 +21,7 @@ def get_sample_content(filename):
             "AVR-X1600H",
             {
                 "friendlyName": "Denon AVR-X1600H",
-                "host": "127.0.0.1",
+                "host": "10.0.0.0",
                 "manufacturer": "Denon",
                 "modelName": "Denon AVR-X1600H",
                 "serialNumber": "XXX",
@@ -32,9 +31,7 @@ def get_sample_content(filename):
 )
 def test_evaluate(model, expected_device):
     """Test that the discovered device looks like expected."""
-    url = "https://127.0.0.1/bar"
-    with requests_mock.Mocker() as mock:
-        mock.get(url, text=get_sample_content(
-            "{model}_upnp.xml".format(model=model)))
-        device = evaluate_scpd_xml(url)
+    url = "https://10.0.0.0/denon"
+    body = get_sample_content("{model}_upnp.xml".format(model=model))
+    device = evaluate_scpd_xml(url, body)
     assert device == expected_device
